// Copyright (c) 2019 WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
//
// WSO2 Inc. licenses this file to you under the Apache License,
// Version 2.0 (the "License"); you may not use this file except
// in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing,
// software distributed under the License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, either express or implied.  See the License for the
// specific language governing permissions and limitations
// under the License.

# Public Ballerina API - Ballerina interface to an AMQP `Channel`.
# To provide AMQ `Channel` related functionalities.
#
# + connection - Reference to a Ballerina RabbitMQ `Connection`.
public type Channel client object {

    Connection? connection;

    # Initializes a Ballerina `Channel` object with the given `Connection` object or connection parameters.
    # Creates a `Connection` object if only the connection configuration is given.
    #
    # + connectionOrConnectionConfig - Holds a Ballerina RabbitMQ `Connection` object or the connection parameters.
    public function __init(ConnectionConfiguration|Connection connectionOrConnectionConfig) {
        self.connection = (connectionOrConnectionConfig is Connection) ?
                                connectionOrConnectionConfig : new Connection(connectionOrConnectionConfig);
        self.createChannel(self.connection);
    }

    # Internally creates an AMQ `Channel`.
    #
    # + connection - Ballerina RabbitMQ Connection object.
    private function createChannel(Connection? connection) = external;

    # Closes this channel.
    #
    public function close() returns error? = external;

    # Actively declare a server-named exclusive, autodelete, non-durable queue or queue with the given configurations.
    #
    # + config - Holds the paramters required to declare a queue.
    # + return - Returns the name of the queue if autogenerated or nil if the queue was successfully generated with the
    #               given parameters. An error is returned if an I/O error is encountered.
    public remote function queueDeclare(QueueConfiguration? queueConfig = ()) returns string | error? = external;

    # Actively declare a non-autodelete, non-durable exchange with no extra arguments, If the arguments are specifed,
    # then the exchange is declared accordingly.
    #
    # + config - Holds parameters required to declare an exchange.
    # + return - Returns an error if an I/O error is encountered or nil if successful.
    public remote function exchangeDeclare(ExchangeConfiguration config) returns error? = external;

    # Binds a queue to an exchange with the given binding key.
    #
    # + queueName - Name of the queue.
    # + exchangeName - Name of the exchange.
    # + bindingKey - Binding key used to bind the queue to the exchange.
    # + return - Returns an error if an I/O error is encountered or nil if successful.
    public remote function queueBind(string queueName, string exchangeName, string bindingKey) returns error? = external;

    # Publishes a message. Publishing to a non-existent exchange will result in a channel-level
    # protocol exception, which closes the channel.
    #
    # + routingKey - The routing key.
    # + message - The message body.
    # + exchange - The name of the exchange the message is published to.
    # + return - Returns an error if an I/O error is encountered or nil if successful.
    public remote function basicPublish(string message, string routingKey,
                                            string exchange = "") returns error? = external;

    # Deletes a queue, without regard for whether it is in use or has messages on it, If the paramters ifUnused or
    # ifEmpty is given, the queue is checked before deleting.
    #
    # + queueName - Name of the queue to be deleted.
    # + return - Returns error if an I/O error is encountered or nil if successful.
    public remote function queueDelete(string queueName) returns error? = external;

    # Deletes an exchange.
    #
    # + exchange - The name of the exchange.
    # + return - An I/O error if an error is encountered or nil otherwise.
    public remote function exchangeDelete(string exchange) returns error? = external;

    # Purges the contents of the given queue.
    #
    # + queueName - The name of the queue.
    # + return - An error if an I/O error is encountered or nil if successful.
    public remote function queuePurge(string queueName) returns error? = external;
};

# Holds the parameters used to declare a queue.
#
# + queueName - The name of the queue, if not specified then autogenerated.
# + durable - True if declaring a durable queue (the queue will survive a server restart).
# + exclusive - True if we are declaring an exclusive queue (restricted to this connection).
# + autoDelete - True if we are declaring an autodelete queue (server will
# delete it when no longer in use).
# + arguments - Other properties (construction arguments) for the queue.
public type QueueConfiguration record {|
    string queueName;
    boolean durable = false;
    boolean exclusive = false;
    boolean autoDelete = true;
|};

# Holds the parameters used to declare an exchange.
#
# + exchangeName - The name of the exchange.
# + exchangeType - The type of the exchange.
# + durable - True if declaring a durable exchange (the exchange will survive a server restart).
public type ExchangeConfiguration record {|
    string exchangeName;
    ExchangeType exchangeType = DIRECT_EXCHANGE;
    boolean durable = false;
|};
